;
;                    Inertia Realtime 3D Rendering Engine
;    Copyright (c) 1996, Alex Chalfin, Jeroen Bouwens. All Rights Reserved.
;          DISTRIBUTION OF THIS SOURCE CODE IS STRICTLY PROHIBITED
;

;
; PLYPHTX.ASM
; code: Alex Chalfin
;

; Phong Texture (affine) polygon filler
;
;================================================================
; Polygon specific defines
;================================================================
;
__INTERP_EXTRA__  EQU 0        ; must be defined for everything but flat
;__INTERP_C__      EQU 0        ; must be defined for goruaud shading
__INTERP_UV__     EQU 0        ; must be defined for texture mapping
__INTERP_ENV__    EQU 0        ; must be defined for env mapping (phong)
;

.486p                           ; 486 protected mode
.Model Flat,Pascal              ; with flat memory model

DATA SEGMENT PARA PUBLIC USE32  ;
                                ;
public PhongTexturePolygonASM   ;
                                ;
include strucdef.asm            ; Inertia structure references
include varsasm.inc             ; external variable references
                                ;
DATA ENDS                       ;
                                ;
;================================================================
; Code Segment
;================================================================
                                ;
CODE SEGMENT PARA PUBLIC USE32  ;
                                ;
;===========================================================================
; Macros
;===========================================================================
                                ;
                                ; All registers are available for loop
                                ; macro. Nothing needs to be saved/restored
                                ; it is all handled in AFFASM.INC
                                ;
INNER_LOOP Macro  Left, Right, Height, Side;
Local  @@HeightLooper,@@ExitHeight,@@SkipFill,@@FillLoop ; Local labels
                                ;
 @@HeightLooper:                ;
  Mov  eax,[Height]             ;
  Cmp  eax,0                    ;
  Jle @@ExitHeight              ;
                                ;
  Dec  [Height]                 ;  decrement the height value by 1
                                ;
  Mov  ebx,[Left&X]             ;
  Add  ebx,65535                ;
  And  ebx,0ffff0000h           ;
                                ;
  Mov  edx,ebx                  ;
  Sub  edx,[Left&X]             ;
                                ; edx contains sub-pixel correction factor
  Mov  esi,edx                  ;
                                ;
  Mov  eax,[HorizontalDU]       ;
  IMul edx                      ;
  Shrd eax,edx,16               ; fix for sub-pixel adjustment
                                ;
  Add  eax,[Side&U]             ; eax is the starting horizontal color
                                ;
  Mov  [UStart],eax             ;
                                ;
  Mov  edx,esi                  ; get sup-pixel correction factor
  Mov  eax,[HorizontalDV]       ;
  IMul edx                      ;
  Shrd eax,edx,16               ; fix for sub-pixel adjustment
                                ;
  Add  eax,[Side&V]             ; eax is the starting horizontal color
                                ;
  Mov  [VStart],eax             ;
                                ;
  Mov  edx,esi                  ; get sup-pixel correction factor
  Mov  eax,[HorizontalDS]       ;
  IMul edx                      ;
  Shrd eax,edx,16               ; fix for sub-pixel adjustment
                                ;
  Add  eax,[Side&S]             ; eax is the starting horizontal color
                                ;
  Mov  [SStart],eax             ;
                                ;
  Mov  edx,esi                  ; get sup-pixel correction factor
  Mov  eax,[HorizontalDT]       ;
  IMul edx                      ;
  Shrd eax,edx,16               ; fix for sub-pixel adjustment
                                ;
  Add  eax,[Side&T]             ; eax is the starting horizontal color
                                ;
  Mov  [TStart],eax             ;
                                ;
  Sar  ebx,16                   ;
                                ;
  Mov  ecx,[Right&X]            ;
  Dec  ecx                      ;
  Sar  ecx,16                   ;
                                ;
  Sub  ecx,ebx                  ;
  Inc  ecx                      ;
  Cmp  ecx,0                    ;
  Jle @@SkipFill                ;
                                ;
  Mov  [Count],ecx              ;
                                ;
  Mov  edi,[YAddress]           ;
  Add  edi,ebx                  ;
                                ;
  Mov  eax,[UStart]             ;
  Mov  ebx,[VStart]             ;
                                ;
  Sar  eax,8                    ;
  Sar  ebx,8                    ;
                                ;
  Mov  [UStart],eax             ;
  Mov  [VStart],ebx             ;
                                ;
  Mov  edx,[HorizontalDU]       ;
  Mov  esi,[HorizontalDV]       ;
                                ;
  Sar  edx,8                    ;
  Sar  esi,8                    ;
                                ;
  Mov  [HDU8],edx               ;
  Mov  [HDV8],esi               ;
                                ;
  Mov  eax,[SStart]             ;
  Mov  ebx,[TStart]             ;
                                ;
  Sar  eax,8                    ;
  Sar  ebx,8                    ;
                                ;
  Mov  [SStart],eax             ;
  Mov  [TStart],ebx             ;
                                ;
  Mov  edx,[HorizontalDS]       ;
  Mov  esi,[HorizontalDT]       ;
                                ;
  Sar  edx,8                    ;
  Sar  esi,8                    ;
                                ;
  Mov  [HDS8],edx               ;
  Mov  [HDT8],esi               ;
                                ;
  Mov  esi,[TStart]             ;
  Mov  ecx,[SStart]             ;
  Mov  eax,[UStart]             ;
  Mov  ebx,[VStart]             ;
                                ; add in the intensity offset
                                ; On entering loop:
                                ; edi = screen position
                                ; esi = 8.8 env T value
                                ; ecx = 8.8 env S value
                                ; eax = 8.8 texture U value
                                ; ebx = 8.8 texture V value
  Push  ebp                     ;
  Mov  ebp,[TextureOfsTable]    ;
 @@FillLoop:                    ;
                                ;
  Mov  ecx,[SStart]             ;
                                ;
  Mov  edx,esi                  ;
  Mov  dl,ch                    ;
  Add  ecx,[HDS8]               ;
  And  edx,0ffffh               ;
                                ;
  Mov  ebp,[IBufferTable]       ;
                                ;
  Add  esi,[HDT8]               ;
  Mov  [SStart],ecx             ;
                                ;
  Mov  ch,[ebp+edx]             ; get phong map color
                                ;
  Mov  ebp,[TextureOfsTable]    ;
                                ;
  Mov  dh,bh                    ;
  Mov  dl,ah                    ;
                                ;
  Add  eax,[HDU8]               ;
  Add  ebx,[HDV8]               ;
                                ;
  And  ecx,0ffffh               ;
                                ;
  Mov  cl,[ebp+edx]             ; texture color
  Mov  ebp,[IntensityTable]     ;
                                ;
  Inc  edi                      ;
                                ;
  Mov  dl,[ebp+ecx]             ; get shade color
                                ;
  Xor  ecx,ecx                  ;
  Dec  [Count]                  ;
                                ;
  Mov  [edi-1],dl               ;
                                ;
  Jnz @@FillLoop                ;
  Pop  ebp                      ;
                                ;
 @@SkipFill:                    ;
                                ;
  Mov  edx,[YAddress]           ; go to the next scanline
  Add  edx,[LineSize]           ;
  Mov  [YAddress],edx           ; YAddress += LineSize
                                ;
  Mov  eax,[Side&S]             ; increment the U span
  Add  eax,[Side&DS]            ;
  Mov  [Side&S],eax             ;
                                ;
  Mov  eax,[Side&T]             ; increment the U span
  Add  eax,[Side&DT]            ;
  Mov  [Side&T],eax             ;
                                ;
  Mov  eax,[Side&U]             ; increment the U span
  Add  eax,[Side&DU]            ;
  Mov  [Side&U],eax             ;
                                ;
  Mov  eax,[Side&V]             ; increment the V span
  Add  eax,[Side&DV]            ;
  Mov  [Side&V],eax             ;
                                ;
  Mov  eax,[Left&X]             ;
  Mov  ebx,[Right&X]            ;
  Add  eax,[Left&DX]            ;
  Add  ebx,[Right&DX]           ;
  Mov  [Left&X],eax             ; LeftX += LeftDX
  Mov  [Right&X],ebx            ; RightX += RightDX
                                ;
  Jmp @@HeightLooper            ;
                                ;
 @@ExitHeight:                  ;
EndM                            ;
                                ;
;===========================================================================
; CullPolys
;===========================================================================
PhongTexturePolygonASM Proc Uses edi eax,Triangle:DWord, VirtualAddress:DWord, LineSize:DWord
                                ;
  Mov edi,[Triangle]            ;
                                ;
  Mov   eax,[edi].Texture       ;
  Mov   [TextureOfsTable],eax   ;  grab and store the texture pointer
                                ;
  Mov   eax,[edi].IBuffer       ;
  Mov   [IBufferTable],eax      ;  grab and store the texture pointer
                                ;
include affasm.inc              ; include affine interpolation
                                ;
  Ret                           ;
                                ;
PhongTexturePolygonASM EndP     ;


CODE ENDS
END


